/*
 * SPDX-FileCopyrightText: 2019 Bruce Cowan <bruce@bcowan.me.uk>
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include "hashtable.h"
#include "mem.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define HASH_SIZE 16

/**
 * HashTable:
 *
 * The #HashTable struct is an opaque data structure to represent a
 * Hash Table. It should only be accessed via the following functions.
 */
struct _HashTable
{
  SList **array;

  HashFunc hash_func;
  EqualFunc equal_func;
  FreeFunc free_key_func;
  FreeFunc free_value_func;
};

typedef struct
{
  void *key;
  void *value;
} KeyValue;

static KeyValue *
key_value_new (const void *key,
               const void *value)
{
  KeyValue *kv = check_malloc (sizeof (KeyValue));

  kv->key = (void *) key;
  kv->value = (void *) value;

  return kv;
}

static void
key_value_free (KeyValue *kv,
                FreeFunc  key_func,
                FreeFunc  val_func)
{
  if (key_func)
    key_func (kv->key);

  if (val_func)
    val_func (kv->value);

  free (kv);
}

/**
 * hash_table_new:
 * @hash_func: A hash function for the key.
 * @equal_func: A function to check for equality.
 * @free_key_func: A function to free keys.
 * @free_value_func: A function to free values.
 *
 * Creates a new #HashTable.
 */
HashTable *
hash_table_new (HashFunc  hash_func,
                EqualFunc equal_func,
                FreeFunc  free_key_func,
                FreeFunc  free_value_func)

{
  HashTable *new = check_malloc (sizeof (HashTable));

  new->array = calloc (HASH_SIZE, sizeof (SList *));
  new->hash_func = hash_func;
  new->equal_func = equal_func;
  new->free_key_func = free_key_func;
  new->free_value_func = free_value_func;

  return new;
}

/**
 * hash_table_free:
 * @self: A #HashTable.
 *
 * Frees the memory taken by a #HashTable.
 */
void
hash_table_free (HashTable *self)
{
  for (size_t i = 0; i < HASH_SIZE; i++)
    {
      SList *bucket = self->array[i];

      // Can't use slist_free all here because of separate key and value frees
      for (SList *l = bucket; l; l = l->next)
        {
          KeyValue *kv = l->data;
          key_value_free (kv, self->free_key_func, self->free_value_func);
        }
    }
}

/**
 * hash_table_lookup:
 * @self: A #HashTable.
 * @key: A key to look up.
 *
 * Finds the value associated with @key.
 *
 * Returns: The value.
 */
const void *
hash_table_lookup (HashTable  *self,
                   const void *key)
{
  unsigned idx = self->hash_func (key) % HASH_SIZE;
  SList *bucket = self->array[idx];

  for (SList *el = bucket; el; el = el->next)
    {
      KeyValue *kv = el->data;
      if (self->equal_func (key, kv->key))
        return kv->value;
    }

  return NULL;
}

/**
 * hash_table_insert:
 * @table: a #HashTable.
 * @key: the key to insert.
 * @value: the value to associate with the key.
 *
 * Inserts a new key and value into a #HashTable.
 *
 * Duplicates are ignored, @key and @value are not copied.
 */
void
hash_table_insert (HashTable  *table,
                   const void *key,
                   const void *value)
{
  unsigned idx = table->hash_func (key) % HASH_SIZE;

  KeyValue *kv = key_value_new (key, value);
  SList *list = slist_prepend (table->array[idx], kv);
  table->array[idx] = list;
}

/**
 * hash_table_foreach:
 * @table: a #HashTable.
 *
 * Runs a function on the contents of a #HashTable.
 */
void
hash_table_foreach (HashTable   *self,
                    HashIterFunc func,
                    void        *user_data)
{
  for (size_t i = 0; i < HASH_SIZE; i++)
    {
      SList *bucket = self->array[i];

      for (SList *el = bucket; el; el = el->next)
        {
          KeyValue *kv = el->data;
          func (kv->key, kv->value, user_data);
        }
    }
}
